/*
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (define.C)
  
  Copyright:  Damian J. J. Farnell

  Date:       Wed Sep 29 1999
 
  Purpose:    Determine and solve CCM LSUBm equations.

  Updated:    Joerg Schulenburg 2005

  ===================================================

  ---------------------------------------------------

  MODULE NAME: DEFINE.C 

  PURPOSE:     Defines the LSUBm clusters recursively.
               Not optimised. 

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

//#include <iostream.h> /* gcc doesn like this, its obsolete */
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

#include "Constants.h"
#include "Externals.h"
#include "FunctionPrototypes.h"

#define MAX_out 100


int subsetClusters(int (*config)[5], int length, int id);
int numSubClusters(int (*config)[5], int length);
void updateDescriptors(int *descriptor, struct spinConfigs *dTable,
                       int numberElements);

// replaced JS200912
void findSpinConfigurations(struct spinConfigs *cfgs, int *nCfgs, int id);
// void findFundamentalFundamentalSpinConfigurations(void);
// void findExcitedFundamentalSpinConfigurations(void);

struct spinConfigs *latticeAnimals = NULL;
int           NumberLatticeAnimals;

// JS 2009-07: new function,
//   called by define() to output animals, fundamental and excited configs
// JS09: output does not work for animals m>n, because Animals.n is set to m
void printOutStoredConfigs(const char *name,
                           struct spinConfigs *spinConfigurations,
                           int NumberConfigurations) {
  
  int i, j, length, outlen, NumberDisconnected=0, NumberConnected=0,
      (*configuration)[5];
  /* as an overview we count 1D,2D,3D and 0..length animals */
  int count[MAX_LSUBM+1][3], sum[3], dim, xyz;
  for(j=0;j<MAX_LSUBM+1;j++) count[j][0]=count[j][1]=count[j][2]=0;
  sum[0]=sum[1]=sum[2]=0;
  configuration = (int (*)[5]) malloc((MAX_LSUBM+1)*5*sizeof(int));

  {
    const char *dims[8]={".  ","x  ", "y  ","xy ","z  ","xz ","yz ","xyz"};
    int nc=0, outidx=0, maxout=0;
    if (Control.output>0) maxout=20;
    if (Control.output>1) maxout=50;
    if (Control.output>2) maxout=100;
    if (Control.output>3) maxout=500;
    /* output sorted by length */
    // (sample: flips=8, maxdist=4) old=Params.m Jun08
    // but Animals flips=2 maxdist=10 ? Dec09
    if (Control.output>0)
    for (outlen=1; outlen<=Params.n; outlen++)
    for(i=1;i<=NumberConfigurations; i++) {
      // length means table-of-flips length = num flips
      length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
                                       configuration);
      if (length != outlen) continue;
      outidx++;
      if ((!myid) && (outidx<maxout || outidx>NumberConfigurations-maxout)) {
          xyz=0; dim=1;
          for (j=1; j<=length; j++) {
            if (configuration[j][1]) xyz|=1;
            if (configuration[j][2]) xyz|=2;
            if (configuration[j][3]) xyz|=4;
          }
          if (xyz==7)                     dim=3; // three bits set
          if (xyz==3 || xyz==5 || xyz==6) dim=2; // two bits set
          if (xyz<=1 || xyz==2 || xyz==4) dim=1; // include Dim=0
          // number of clusters which are disconnected from the others
          nc = numSubClusters(configuration, length);
          // most compact one set per line
          printf("Cluster %6d (parts=%d dim=%s): ", i, nc, dims[xyz]);
          printOutClusterFromDescriptor(spinConfigurations[i].descriptor);
          if (outidx==maxout-1 && outidx<=NumberConfigurations-maxout)
            printf("  ... skip output ...\n");
      } // if(!myid)
    }
    /* count for statistics */
    for(i=1;i<=NumberConfigurations; i++) {
      length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
                                       configuration);
      xyz=0; dim=1;
      for (j=1; j<=length; j++) {
        if (configuration[j][1]) xyz|=1;
        if (configuration[j][2]) xyz|=2;
        if (configuration[j][3]) xyz|=4;
        if (xyz==7) break;
      }
      if (xyz==7)                     dim=3; // three bits set
      if (xyz==3 || xyz==5 || xyz==6) dim=2; // two bits set
      if (xyz<=1 || xyz==2 || xyz==4) dim=1; // include Dim=0
      // length <= MAX_LSUBM ??? better max_flips?
      if (length>MAX_LSUBM) ERR("n>MAXLSUBM"); /* ToDo: need max_n? */
      count[length][dim-1]++;
      // number of clusters which are disconnected from the others
      nc = numSubClusters(configuration, length);
      if (nc>1) NumberDisconnected++;
          else  NumberConnected++;
    }
  }
  if (!myid) {
    printf("number_of_clusters (%s):\n"
           " length |     1D       2D         3D |     1D+2D+3D\n"
           "--------+----------------------------+-------------\n", name);
    for (i=1;i<=Params.n;i++) {  
      sum[0]+=count[i][0];
      sum[1]+=count[i][1];
      sum[2]+=count[i][2];
      if (count[i][0]+count[i][1]+count[i][2]) // JS0910 print only nonzeros
        printf(" %6d | %6d %8d %10d | %12d\n",
             i,count[i][0],count[i][1],count[i][2],
               count[i][0]+count[i][1]+count[i][2]);
    }
    printf("--------+----------------------------+-------------\n"
           "   Sum: | %6d %8d %10d | %12d\n",
             sum[0], sum[1], sum[2], sum[0]+sum[1]+sum[2]);
    printf("\n");
    printf("Number of Connected Clusters    = %8d (bad for tri_f1, ToDo)\n", NumberConnected);
    printf("Number of Disconnected Clusters = %8d\n", NumberDisconnected);
    printf("Total Number of Clusters        = %8d\n", NumberConfigurations);
    printf("\n");
  }
}


void define(void)
{

  /* this module defines all of the cluster configurations using
     a recursive algorithm. 

     In essence this is a souped-up version of the algorithm used
     in the original ``define configurations'' C program. Here, for
     complex crystallogrphic lattices and higher spin.

  */

  int i, tempLatticeSymmetries;
  int cell;
  char configurations[80];
  int (*configuration)[5],  (*tempConfig)[5];
  int length;

  tempConfig    = (int (*)[5]) malloc((MAX_LSUBM+1)*5*sizeof(int));
  configuration = (int (*)[5]) malloc((MAX_LSUBM+1)*5*sizeof(int));

  latticeAnimals =
    (struct spinConfigs *) malloc((MAX_NUM_CONFIGURATIONS+1)
                                  *sizeof(struct spinConfigs));
  if (!latticeAnimals) {
    printf("ERROR: malloc latticeAnimals[MAX_NUM_CONFIGURATIONS]\n");
    MPI_Abort(MPI_COMM_WORLD,1); 
  };
  sprintf(configurations, "%s.configs", Control.name);
  
  if (Control.output > 0 && myid == 0) {
    printf("\n============================================================\n");
    printf( "             DEFINE CLUSTERS: START ALGORITHM.\n");
    printf("============================================================\n\n");
    printf("DBG: MAX_NUM_CONFIGURATIONS= %ld\n", (long)MAX_NUM_CONFIGURATIONS);
    printf("DBG: DESCR_SIZE= %d\n", DESCR_SIZE);
  }

  fflush(stdout);

  /* !! NEW ALGORITHM FOR S>= 1/2.... 30-8-2000 */

  /* PART 1: DETERMINE THE LATTICE ANIMALS */
  /* ToDo: store only the biggest, which are not covered by others?
   * because clusters are always parts of the animals (clusters<=animals)
   */

  NumberLatticeAnimals=0;
  NumberConfigurations=0;

  tempLatticeSymmetries=LatticeSymmetries.number;
  //  LatticeSymmetries.number=1; // switch of symmetries, if not faster Oct05
  // can be done parallel?

  if (myid == 0)
    printf("PART 1: DETERMINE THE LATTICE ANIMALS (using %d symmetries)\n\n",
          LatticeSymmetries.number);

  // ToDo: Unitcell members could be introduced later like s
  //       so animals would live only on unitcells and depend from
  //       the lattice and its symmetry?
  // JS08: failes if dimension is 1 but neighbours in y direction (lc_j1j2j3)
  for(cell=1;cell<=UnitCell.number;cell++) {
    length=1;
    configuration[1][1]=0;
    configuration[1][2]=0;
    configuration[1][3]=0;
    configuration[1][4]=cell;      // start site = 1..UC
    /* this includes DSUB-filtering, connected */
    recursiveClusteringFunction(configuration, length,
                        Params.m, // ToDo: old Params.m, new=.n ??? Sep05
			&NumberLatticeAnimals);
  }
  if (myid == 0)
    printf("\ndefine: number of lattice animals: %4d\n", NumberLatticeAnimals);


  if (Control.output > 0 && myid == 0) {
    /* PRINT OUT INFORMATION ON ANIMALS */
    /* as an overview we count 1D,2D,3D and 0..length animals */
    printOutStoredConfigs("lattice animals",
                       latticeAnimals, NumberLatticeAnimals);
  }
   
  if (!myid)
   printf("\nDEFINE LATTICE ANIMALS: COMPLETED SUCCESSFULLY.\n\n");

  {  struct spinConfigs *temp;
     temp =
       (struct spinConfigs *) realloc(latticeAnimals,
                            (NumberLatticeAnimals+1)*sizeof(struct spinConfigs));
     if (temp) latticeAnimals = temp;
     // realloc left memory untached if it fails
  }


  spinConfigurations =
    (struct spinConfigs *) malloc((MAX_NUM_CONFIGURATIONS+1)
                                 *sizeof(struct spinConfigs));
  if (!spinConfigurations) {
    printf("ERROR: malloc spinConfigurations[MAX_NUM_CONFIGURATIONS]\n");
    MPI_Abort(MPI_COMM_WORLD,1); 
  };

  /* PART 2: DETERMINE THE FUNDAMTENTAL SPIN CONFIGURATIONS */
  
  if (!myid)
   printf("\nPART 2: DEFINE FUNDAMENTAL SPIN CONFIGURATIONS"
          " (covered by animals).\n\n");

  /* definition of configurations: */
  /* configurations are  animals who's parts are weigthed by 0..2*s */

  LatticeSymmetries.number=tempLatticeSymmetries;
  // select 0=fundamental, 1=excited
  findSpinConfigurations(spinConfigurations, &NumberConfigurations, 0);

  {  struct spinConfigs *temp;
     temp =
       (struct spinConfigs *) realloc(spinConfigurations,
                            (NumberConfigurations+1)*sizeof(struct spinConfigs));
     if (temp) spinConfigurations = temp;
     // realloc left memory untached if it fails
  }
 
  if (!myid)
   printf("\nDEFINE FUND. SPIN CONFIGS.: COMPLETED SUCCESSFULLY.\n\n");

  /* PART 3: PRINT OUT FUNDAMENTAL SPIN CONFIGURATIONS */
  
  if (!myid)
    printf("\nPART 3: PRINT OUT FUNDAMENTAL SPIN CONFIGURATIONS.\n\n"); 

  printOutStoredConfigs("fundamental spin configs",
                        spinConfigurations, NumberConfigurations);

  // JS Apr08: write only myid==0, distribute per MPI (faster)
  if (!myid)
  { // file format version 2006-06-11 JS
    // v060611: to have easier filtering using dd bs=.. skip=.. remove Num
    // pre060611: fwrite(&NumberConfigurations,sizeof(int),1,fp1);
    printf("Write ConfigDescriptors(%d*%d*(int=%d)) to %s\n\n",
      NumberConfigurations, DESCR_SIZE, (int)sizeof(int), configurations);
    fp1=fopen(configurations,"w");   // $(Control.name).configs
    if (!fp1) { printf("define: Open failed...\nExiting!!\n"); ERR(1); }

    for(i=1;i<=NumberConfigurations;i++) {
      fwrite(spinConfigurations[i].descriptor,
             ((DESCR_SIZE)*sizeof(int)), 1, fp1);
    }
    fclose(fp1);
  }


  if (Control.excitations) 
{
  excitedSpinConfigurations =
    (struct spinConfigs *) malloc((MAX_NUM_CONFIGURATIONS+1)*sizeof(struct spinConfigs));
  if (!excitedSpinConfigurations) {
    printf("ERROR: malloc excitedSpinConfigurations[MAX_NUM_CONFIGURATIONS]\n");
    MPI_Abort(MPI_COMM_WORLD,1); 
  }
  findSpinConfigurations(excitedSpinConfigurations, &NumberExcitedConfigurations, 1);

  if (!myid)
    printf("\nPART 3b: PRINT OUT FUNDAMENTAL EXCITED-STATE SPIN CONFIGURATIONS.\n\n"); 
  
  printOutStoredConfigs("fundamental excited-state spin configs",
                     excitedSpinConfigurations, NumberExcitedConfigurations);

  // JS Apr08: write only myid==0, distribute per MPI (faster)
  if (!myid)
  { // file format version 2006-06-11 JS
    // v060611: to have easier filtering using dd bs=.. skip=.. remove Num
    // pre060611: fwrite(&NumberConfigurations,sizeof(int),1,fp1);
    sprintf(configurations, "%s.configs_exc", Control.name);
     printf("Write Excited-State ConfigDescriptors(%d*%d*(int=%d)) to %s\n\n",
      NumberExcitedConfigurations, DESCR_SIZE, (int)sizeof(int), configurations);
      fp1=fopen(configurations,"w");   // $(Control.name).configs
      if (!fp1) { printf("define: Open failed...\nExiting!!\n"); exit(1); }

      for(i=1;i<=NumberExcitedConfigurations;i++) {
        fwrite(excitedSpinConfigurations[i].descriptor,
             ((DESCR_SIZE)*sizeof(int)), 1, fp1);
      }
      fclose(fp1);
  }

  if (!myid)
    printf("DEFINE EXCITED-STATE CLUSTERS: ALGORITHM SUCCESFULLY COMPLETED!\n\n");

  {  struct spinConfigs *temp;
     temp =
       (struct spinConfigs *) realloc(excitedSpinConfigurations, 
                  (NumberExcitedConfigurations+1)*sizeof(struct spinConfigs));
     if (temp) excitedSpinConfigurations = temp;
     // realloc left memory untached if it fails
  }

}

  if (!myid)
    printf("DEFINE CLUSTERS: ALGORITHM SUCCESFULLY COMPLETED!\n\n");

  free (latticeAnimals);  // may be lot of memory freed now
  latticeAnimals = NULL;

}

// Oct05: new function numSubClusters by Joerg
int numSubClusters(int (*config)[5], int length) {

  int flag_changed=0; // flag if nn-matrix is initialized
  int cluster[MAX_LSUBM+1]; // ToDo: I mean SUBN
  int i, j, k, i2, index, num_clusters=length, tmp[5];

  if (length>MAX_LSUBM) { printf("Error: length>MAX_LSUBM\n"); ERR(1); }
   
  for (i=1; i<=length; i++) cluster[i]=i; // start with each site as cluster

  // ToDo: because of recursiv creation of animals better coding would be
  //       pointer to predecissor + added site
  //       tables[configs]={backpointer+(x,y,z,uc)} and n-steps for decoding!
  for (;;) {
    flag_changed=0; // reset flag
    for(i=1;i<=length-1;i++) {
      index = config[i][4];  // cell
      for(k=i+1; k<=length; k++)
        if (cluster[k]!=cluster[i]) // only different clusters
        if (config[i][1] == config[k][1] &&
            config[i][2] == config[k][2] &&
            config[i][3] == config[k][3] &&
            config[i][4] == config[k][4]) {
        num_clusters--; flag_changed=cluster[k];
        for (i2=1;i2<=length;i2++)
          if (cluster[i2]==flag_changed) { cluster[i2]=cluster[i]; }
      }
      for(j=1;j<=LatticeNeighbours.number[index];j++) {
	tmp[0]=config[i][1]+LatticeNeighbours.x[index][j];
	tmp[1]=config[i][2]+LatticeNeighbours.y[index][j];
	tmp[2]=config[i][3]+LatticeNeighbours.z[index][j];
	tmp[3]=             LatticeNeighbours.cell[index][j];
	
	// check that we have a given site.
	for(k=i+1; k<=length; k++)
          if (cluster[k]!=cluster[i]) // only different clusters
	  if (tmp[0] == config[k][1] &&
	      tmp[1] == config[k][2] &&
	      tmp[2] == config[k][3] &&
	      tmp[3] == config[k][4]) {
            num_clusters--; flag_changed=cluster[k];
            for (i2=1;i2<=length;i2++)
              if (cluster[i2]==flag_changed) { cluster[i2]=cluster[i]; }
        }
      }
    }
    if (!flag_changed) break;
  } // until no changes
  return num_clusters;
}

// Dec09: new function maxDistance from site 0 by Joerg
//  set site 1 to dist 1, rest m+1
//  search all neighbours of each site for min dist and set to dist +1
//    repeat if any distance has changed
int maxDistance(int (*config)[5], int length) {

  int flag_changed=0; // flag if nn-matrix is initialized
  int dist[MAX_LSUBM+1]; // ToDo: I mean SUBN
  int i, j, k, i2, index, maxD=MAX_LSUBM+1, tmp[5];

  if (length<2) return length;
  if (length>MAX_LSUBM) { printf("Error: length>MAX_LSUBM\n"); ERR(1); }
   
  for (i=1; i<=length; i++) dist[i]=MAX_LSUBM+1; // start with each site m+1
  dist[1]=1; // must be xyc=0 0 1

  for (;;) {
    flag_changed=0; // reset flag
    for(i=1;i<=length-1;i++) {
      index = config[i][4];  // cell
      i2=MAX_LSUBM+1;
      for(j=1;j<=LatticeNeighbours.number[index];j++) {
	tmp[0]=config[i][1]+LatticeNeighbours.x[index][j];
	tmp[1]=config[i][2]+LatticeNeighbours.y[index][j];
	tmp[2]=config[i][3]+LatticeNeighbours.z[index][j];
	tmp[3]=             LatticeNeighbours.cell[index][j];
	
	// check that we have a given site.
	for(k=i+1; k<=length; k++)
	  if (tmp[0] == config[k][1] &&
	      tmp[1] == config[k][2] &&
	      tmp[2] == config[k][3] &&
	      tmp[3] == config[k][4]) {
            if (dist[k]+1 < dist[i]) {
              dist[k]=dist[i]+1; flag_changed=1;
            }
        }
      }
    }
    if (!flag_changed) break;
  } // until no changes
  for (i=1; i<length; i++) if (maxD>dist[i]) maxD=dist[i];
  return maxD;
}



// in:   configuration=matrix(0,MAX_LSUBM+1,0,5)={0,0,0,1} initially
//       length=1 initially
//       maxorder=Params.m=m(SUBn-m) distance?
// work: configuration=matrix(0,MAX_LSUBM+1,0,5)
// out:  latticeAnimals[].descriptor via updateDescriptors
//       NumberAnimals  (=NumberLatticeAnimals)
  // ToDo: because of recursiv creation of animals better coding would be
  //       pointer to predecissor + added site
  //       tables[configs]={backpointer+(x,y,z,uc)} and n-steps for decoding!
  //   do it parallel!
void recursiveClusteringFunction(int (*configuration)[5], 
		 int length, int maxorder, int *NumberAnimals) {

  /* A simple recursive function that determines all connected configs */

  int i,j,k, descriptorIndex;
  int index, tmp[4], err;
  static int descriptor[DESCR_SIZE];
  static int n_out=0;
  
  // first check whether we have an error using the `length' of cluster. 

  if (length > maxorder) {
    printf("Error in Recursion: length > maxorder!\nExiting...\n\n");
    ERR(1); 
  } 

  // first check if the configuration has been seen before
  
  // get descriptor from symmetrized configuration[tempLength]
  // without symmetry: findConfigurationToDescriptor(descriptor, configuration, length);
  //  see bookKeeping.C
//  if (!subsetClusters(configuration,length,0)) err=-1; else // JS0912 ???
  err = findDescriptorOfCluster(length, configuration, descriptor);
  if (Control.output > 2 && myid == 0 && *NumberAnimals < 100) {
    printf("Test %d  Animal    : ", err); 
    printOutConfig(configuration, length);
  }

  if (err == -1) { // ??? Joerg Oct05 Oct09
    printf("Error recursivClustering: invalid descriptor\n");
    ERR(1);  // DESCR_SIZE to small?
  }
#if 0 // JS0912
  if (err==0) return; // testwise added JS0912 (is not min)
#endif

  // ToDo: problems on NumberAnimals MAX_NUM_CONFIGURATIONS-overflow
  descriptorIndex = findDescriptorIndex(descriptor, latticeAnimals, *NumberAnimals);
  
  // 0 means "not yet stored"
  if (Control.output > 3 && myid == 0 && *NumberAnimals < 100)
    printf("previously stored as = %d (0=not stored)\n\n", descriptorIndex);

  // this is new and only works together with the other changes of Oct05
  if (descriptorIndex) return; // much faster if symmetries are used, Oct05
  
  // allow also length<=maxorder (old was: length==maxorder) Oct05 (faster)
  if ((!descriptorIndex) && (length <= maxorder)) {
    // store descriptor to latticeAnimals[].descriptor
    updateDescriptors(descriptor, latticeAnimals, *NumberAnimals);
    (*NumberAnimals)++;
    if (Control.output > 1 && myid == 0 && n_out < 100) {
      n_out++;
      // most compact one set per line
      printf("Lattice Animal %3d: ", *NumberAnimals);
      printOutClusterFromDescriptor(descriptor);
    }
  }

  
  // ToDo: if (length == Params.n) return; // maxorder == Params.n?
  if (length == maxorder) {

    // this is a `dummy case' which just makes sure that we don't
    // go over the maximum LSUBm approximation number adjacent
    // lattice sites for the lattice animals.

  } else {

    // length = 0...MAX_LSUBM, ToDo: check outside maxorder <= MAX_LSUBN
    if (length > MAX_LSUBM) {
       printf("Error overflow: length=%d > MAX_LSUBM\n", length);
       ERR(1);
    }
    for(i=1;i<=length;i++) {
      index = configuration[i][4];  // unit cell site index
      if (Control.output > 4 && myid == 0 && *NumberAnimals < 100)
        printf("Check %d Neighbours of site(%d):\n",
               LatticeNeighbours.number[index], i);
      for(j=1;j<=LatticeNeighbours.number[index];j++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.
	tmp[0]=configuration[i][1]+LatticeNeighbours.x[index][j];
	tmp[1]=configuration[i][2]+LatticeNeighbours.y[index][j];
	tmp[2]=configuration[i][3]+LatticeNeighbours.z[index][j];
	tmp[3]=                 LatticeNeighbours.cell[index][j];
	
        // if (tmp[0] < 0) continue; // removed for the new version
//printf(" DBG: try nn %2d %2d %d\n", tmp[0], tmp[1], tmp[3]);
	
	// check that we have NOT repeated a given site.
	for(/*numberFlips=1,*/k=1;k<length+1;k++)
	  if (tmp[0] == configuration[k][1] &&
	      tmp[1] == configuration[k][2] &&
	      tmp[2] == configuration[k][3] &&
	      tmp[3] == configuration[k][4]) break;
        if (k<length+1) continue;

#if 0
        // JS0912 check that the new site is in the BT
        if (tmp[2]<0 || tmp[1]<0 || tmp[0]<0) continue;
        // JS0912 check that the new site index is greater
        if (tmp[2]< configuration[length][3]) continue;
        if (tmp[2]==configuration[length][3]
         && tmp[1]< configuration[length][2]) continue;
        if (tmp[2]==configuration[length][3]
         && tmp[1]==configuration[length][2]
         && tmp[0]< configuration[length][1]) continue;
        if (tmp[2]==configuration[length][3]
         && tmp[1]==configuration[length][2]
         && tmp[0]==configuration[length][1]
         && tmp[3]<=configuration[length][4]) continue;
#endif

#ifdef DSUB 
        // this is an extension suggested by Raymond Bishop and Peggy Li
        // only Clusters with a maximum distance of L(m)^2 is allowed
        //  n DSUBm L^2 animals configs  Eg/N    (sample square Jan06)
        //  2   1   1    2       1       -0.64833148  8xMIPS-250MHz 
        //  4 2=3   2-4  4       2       -0.65311289  8x1s
        //  6   4   5    17      12      -0.66258044  8x5s
        //  8   5   8    34      20      -0.66307028  1x110s terms=2431     
        //  10  6   9    48      43      -0.66511173  1x222s 8obtain=15..93s
        //  12  7   10   170     135     -0.66589305  8x12934s=50t6         
        //  14  8   13   703     831     -0.66703559  62s++ =36t7 (E to high?) 
        //  16  9   16   963     1225    -0.6670114795 1+89s++ =65t7  E9>E8 ???    
        //      8xR10k-250MHz-CPUs: 17..90h sum=254h=8*31.7h efficients=0.35
        //  18  10  17   3620    6874                 61+1147s++               
        //  20  11  18   6827    14084                199+3226s++
        //   m=n bad? better n=m^2 ToDo: what about n!=m?
        // plot [:] "a" u ($5**-.5):7 w lp
        // connectedConfigs  <= Animals (1s) but disconnected can be greater
        static double max_dist2=-1, dist2, x1, y1, z1, x2, y2, z2;
        // m is used otherwise, so we redefine it in a useful manner
        const int L2[26]={0,1,1,2,2,4,5,5,8,8,9,9,10,10,13,13,16,16,
                          17,17,18,18,20,20};
               // L(m/2=1..7) replaced later
        if (Params.m > 25) {
          printf("L(m) not defined for m>25. Abort.\n"); ERR(1); }
        if (max_dist2 < 0 && myid == 0)
          printf(" ==> using DSUBm filter algorithm, L^2=%d <==\n",
                 L2[Params.m]);
        findCartesiansFromGeneralisedPosition(&x1,&y1,&z1,tmp-1);
        for (k=1;k<length+1;k++) {
          findCartesiansFromGeneralisedPosition(&x2,&y2,&z2,configuration[k]);

          dist2 = (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1);

          if (k==1 || (k>1 && max_dist2<dist2)) max_dist2=dist2;
        if (Control.output > 2 && myid == 0 && n_out < 100)
          printf(" k = %2d  dist^2 = %2g  max^2 = %2g"
                 "  %2g %2g %2g  %2g %2g %2g\n",
           k, dist2, max_dist2, x1, y1, z1, x2, y2 ,z2);
        }
        if (max_dist2 > L2[Params.m]+0.1) continue;
        if (Control.output > 2 && myid == 0 && n_out < 100)
          printf(" maxdist^2 = %g, animal belongs to DSUB\n", max_dist2);
#endif

#ifdef LDSUB 
        // generating bigger LSUB and filter smaller LDSUB is slow
        //  ToDo: better generating LSUB directly
        // this is an extension suggested by Raymond Bishop and Peggy Li
        // only Clusters with a maximum Manhatten distance of L(m) 
        //  n LDSUBm |L| animals configs  Eg/N    (sample square Jan06)
        //  2   1   1    2       1       -0.64833148  8xMIPS-250MHz 
        //  4   2   2    -       2
        //  8   3   3            9 <=> 25 ???
        // 24   4   4           # n=((L+1)*(L+1)) m=L*(L+1)+L ???
        // For example this (biggest) cluster should belong to LDSUB3:
        // .oo.    . = empty places
        // oooo    o = belong to the animal/config
        // oooo
        // .oo.
        static double max_dist2=-1, dist2, x1, y1, z1, x2, y2, z2;
        // m is used otherwise, so we redefine it in a useful manner
        // m(=n?) is min num steps to create config ???
        const int L2[26]={0,1,1,1,2,2,2,2,3,3,3,3,3,3,3,3,3,3,
                          18,19,20,21,22,23,24,25};
        if (Params.m > 25) {
          printf("L(m) not defined for m>25. Abort.\n"); ERR(1); }
        if (max_dist2 < 0 && myid == 0)
          printf(" ==> using LDSUBm filter algorithm, L=%d <==\n",
                 L2[Params.m]);
        findCartesiansFromGeneralisedPosition(&x1,&y1,&z1,tmp-1);
        for (k=1;k<length+1;k++) {
          findCartesiansFromGeneralisedPosition(&x2,&y2,&z2,configuration[k]);

          // using manhattan distance
          dist2 = fabs(x2-x1) + fabs(y2-y1) + fabs(z2-z1);

          if (k==1 || (k>1 && max_dist2<dist2)) max_dist2=dist2;
          if (Control.output > 2 && myid == 0 && n_out < 100)
          printf(" k = %2d  dist = %2g  max = %2g"
                 "  %2g %2g %2g  %2g %2g %2g\n",
           k, dist2, max_dist2, x1, y1, z1, x2, y2 ,z2);
        }
        if (max_dist2 > L2[Params.m]+0.1) continue;
        if (Control.output > 2 && myid == 0 && n_out < 100)
          printf(" maxdist = %g, animal belongs to LDSUB\n", max_dist2);
#endif

#ifdef LPSUB 
        // generating bigger LSUB and filter smaller LPSUB is slow
        //  ToDo: better generating LPSUB directly
        // this is an extension suggested by Raymond Bishop and Peggy Li
        // only Clusters within an x*y*z array with x+y+z <= L(m)
        //  n LDSUBm |L| animals configs  Eg/N    (sample square Jan06)
        //  2   1   1    2       1       -0.64833148  8xMIPS-250MHz 
        //  4   2   2    -       2
        //  6   3   3            9
        //  8   4   4
        // 12   5   5
        // 16   6   6  # max 3 steps in x and y direction,
        // 20   7   7  # n=((L+1)/2+1)*(L/2+1)  m=n-1
        // For example this (biggest) cluster should belong to LPSUB3:
        // 3x0      2x1
        // oooo  OR ooo    . = empty places
        //          ooo    o = belong to the animal/config
        static double max_dist2=-1, max_x2, max_y2, max_z2, dist2, x1, y1, z1, x2, y2 ,z2;
        // m is used otherwise, so we redefine it in a useful manner
        // m(=n?) is min num steps to create config ???
        const int L2[26]={0,1,1,1,2,2,3,3,4,4,4,4,5,5,5,5,6,6,
                          6,6,7,7,7,7,7,8};
        if (Params.m > 25) {
          printf("L(m) not defined for m>25. Abort.\n"); ERR(1); }
        if (max_dist2 < 0 && myid == 0)
          printf(" ==> using LDSUBm filter algorithm, L=%d <==\n",
                 L2[Params.m]);
        findCartesiansFromGeneralisedPosition(&x1,&y1,&z1,tmp-1);
        max_x2=x1; max_y2=y1; max_z2=z1;
        max_dist2 = fabs(max_x2) + fabs(max_y2) + fabs(max_z2);
        for (k=1;k<length+1;k++) {
          findCartesiansFromGeneralisedPosition(&x2,&y2,&z2,configuration[k]);

          if (x2>max_x2) max_x2 = x2;
          if (y2>max_y2) max_y2 = y2;
          if (z2>max_z2) max_z2 = z2;

          max_dist2 = fabs(max_x2) + fabs(max_y2) + fabs(max_z2);

          if (Control.output > 2 && myid == 0 && n_out < 100)
          printf(" k = %2d  dist = %2g  max = %2g"
                 "  %2g %2g %2g  %2g %2g %2g\n",
           k, dist2, max_dist2, x1, y1, z1, x2, y2 ,z2);
        }
        // using manhattan distance
        if (max_dist2 > L2[Params.m]+0.1 || x1<0 || y1<0 || z1<0) continue;
        if (Control.output > 2 && myid == 0 && n_out < 100)
          printf(" max_dist = %g, animal belongs to LPSUB\n", max_dist2);
#endif

        // add new neighbouring site to the lattice animal
	configuration[length+1][1] = tmp[0];
	configuration[length+1][2] = tmp[1];
	configuration[length+1][3] = tmp[2];
	configuration[length+1][4] = tmp[3];
	
        // now do the recursion
	recursiveClusteringFunction(configuration, length+1, maxorder,
				      NumberAnimals);
      }
    }
  }
}

// Joerg Sep05: simplify code
// compare descriptors
// return: 1 desc1>desc2, 0 desc1==desc2, -1 desc1<desc2
int cmpDescriptors(int *desc1, int *desc2) {
  int i;
  for(i=DESCR_SIZE-1;i>=0;i--) {
    if (desc2[i]<desc1[i]) { return  1; } else // greater
    if (desc2[i]>desc1[i]) { return -1; }      // smaller
  }
  return 0; // equal
}

// added by Joerg (faster than animals reduction detour)
// out: NumberConfigurations, ClusterFile.terms
// called by define() - DEFINE FUNDAMENTAL SPIN CONFIGURATIONS
// old -> see findFundamentalSpinConfigurations(void)
//   start with findRecursiveConfigs(0,config[0][0]);
// CPU propto (Params.m-1)!
// ToDo: disconnected Clusters are only valid, if they ly inside
//       a connected SUBn-m cluster (n-m animal with holes)
//       until know they are counted too (to much for D>1)
// ToDo: symmetric m-animals + select n from mm<=m flips (n>m valid for s=1)
//       or at sites with (x y z uc) greater last site!
//  sample: oo.o is invalid LSUB4, but its valid LSUB5+LSUB6
//          .o..
//          ....
//  tested:
//    - square           2D bravais    s=1/2 nUC=1 = OK
//    - triangle_field_1 2D nonbravais s=1/2 nUC=3 = FAILED to big (JS09)
//   ToDo: mixed s ? FIX IT (JS0912)
int findRecursiveConfigs(int length, int (*config)[5])  // Joerg Nov05
{
  int i, j, uc, nc=0, flips, dist, is_minDescriptor, is_subset;
  static long int newNumClusters=0, // numConfigs (for the test case)
                  allClusters=0;    // num tested configs (loop length)
  int descriptor[DESCR_SIZE];
  static int n_out=0;

  if (length>Params.n) return 0;
  if (length==0 && !myid) printf("\nfindRecursiveConfigs:\n");
  
  // generate valid configs from input testconfig
  // setting n flips on length sites (first and last must be set!)
  if (length>0)
  for (i=0;i<1;i++) { // just to use break
    // ToDo: better call Subset before calling  slow functions!
    //   call subsetClusters with cluster instead of descriptor!
    is_subset=subsetClusters(config,length,0);  // ToDo: use id=1 for exc.

    // get descriptor from symmetrized config[length]
    //  add 1 if we have already the minimum descriptor
    // extern int flag_subset must be 0!?
    is_minDescriptor= 0;
//    if (is_subset) // JS0912 bad try
// is_min is needed to avoid double recursiv configs also for subset=0
    is_minDescriptor= findDescriptorOfCluster(length, config, descriptor);
    if (is_subset && is_minDescriptor==1) newNumClusters++; // JS0912
    if (is_subset) allClusters++;
    if (is_minDescriptor!=1) return 0; // JS0912
     
    if (Control.output > 2 && myid == 0 && n_out<=100) {
      printf("\n fRC-test %ld min=%d subset=%d: ",
         allClusters, is_minDescriptor, is_subset);
      printOutConfig(config, length);
      if (Control.output > 3) { printf(" transformed into: ");
                                printOutClusterFromDescriptor(descriptor); }
      if (Control.output > 3 && is_minDescriptor==1) // testwise
        printf("  count= %4ld", ((is_minDescriptor==1)?newNumClusters:-1L));
      // printf("\n");
    }
    if (n_out<100000) n_out++; // output counter
        
    if (!is_subset) break;
    if (is_minDescriptor!=1) break;
    // ToDo: also filter out here length > Params.n (spin flips)
    static long nc_out = 1;

    // if (NumberConfigurations < MAX_NUM_CONFIGURATIONS)
    //  updateDescriptors(descriptor,
    //                    spinConfigurations,NumberConfigurations);
    // NumberConfigurations++;
    nc=1;
    
    if (!myid)
    if ((Control.output>2 && NumberConfigurations<=60)
      ||(Control.output>0 && NumberConfigurations==nc_out)) {
      if (NumberConfigurations==nc_out) nc_out<<=1; // log output
//      printf(" add config= %6d: ", NumberConfigurations);
      printf(" add config= %6ld: ", newNumClusters);
      printOutClusterFromDescriptor(descriptor);
      if (NumberConfigurations==60) printf("... skip output ...");
    }
    break;
  }

  // here starts recursive testconfig (animal generation)
  // JS: how are the sites ordered in BTarray?
  //     how are the sites in configFromDescr ordered?
  for (i=BTarray[config[length][3]*Params.btLength
                +config[length][2]]
                +config[length][1]; i<sizeLUT; i++) {
      config[length+1][1]=lookUpTable[i][0];
      config[length+1][2]=lookUpTable[i][1];
      config[length+1][3]=lookUpTable[i][2];
#if 0
      if (lookUpTable[i][0]<0) break;
      if (lookUpTable[i][0]>=Params.m) continue;
#endif
#if 1
      // new site must be bigger than old sites
      if (length>0) { // xyz= 000 100 200 010 110 210 ...
        if (config[length+1][3]< config[length][3]) continue;
        if (config[length+1][3]==config[length][3]
         && config[length+1][2]< config[length][2]) continue;
        if (config[length+1][3]==config[length][3]
         && config[length+1][2]==config[length][2]
         && config[length+1][1]< config[length][1]) continue;
      }
#endif
      for (uc=1; uc<=UnitCell.number; uc++) { 
        config[length+1][4]=uc;
        // count num flips for this site, stop if maximum reached
        // maximum distance must be <= Params.m
        // ToDo: sizeM (defined as size of covering connected cluster) must be <= M
        //       can be calculated by numSubClusters too (or integrate here)
        // slowest part! ToDo: works for kago test (bad distance?)
        //   ToDo: store flips on sites (expand later), no counting needed
        //   distance = distance to (0,0,0)? within triangle?
        //    or ...
        for(flips=0,dist=0,j=1;j<=length;j++) {
          if (config[j][1]==config[length+1][1]
           && config[j][2]==config[length+1][2]
           && config[j][3]==config[length+1][3]
           && config[j][4]==config[length+1][4]) flips++;
          // JS0912 ToDo: distance does not work for UnitCell > 1
          if (abs(config[j][1]-config[length+1][1])>dist) dist=
              abs(config[j][1]-config[length+1][1]);
          if (abs(config[j][2]-config[length+1][2])>dist) dist=
              abs(config[j][2]-config[length+1][2]);
          if (abs(config[j][3]-config[length+1][3])>dist) dist=
              abs(config[j][3]-config[length+1][3]);
        }
//        dist= maxDistance(config,length+1)-1;  // JS0912
        if (flips>=UnitCell.s2[uc]) continue;
        if (dist >=Params.m       ) continue;
#if 1
        if (length>0)
        if (config[length+1][1]==config[length][1]
         && config[length+1][2]==config[length][2]
         && config[length+1][3]==config[length][3]
         && config[length+1][4]<=config[length][4]) continue;
#endif
//      printf("DBG: rec l=%d i=%d xyc= %d %d %d\n", length,i,lookUpTable[i][0],lookUpTable[i][1],uc);
        nc += findRecursiveConfigs(length+1, config);
      }
   }
   return nc;
}

// in:  lattice animals (only connected clusters)
//      id: 0 = FundamentalConfigurations
//          1 = ExcitedConfigurations     JS0912
// out: ClusterFile.terms, *cfgs, *nCfgs
// called by define() - DEFINE FUNDAMENTAL SPIN CONFIGURATIONS
//    - we remove parts of the animals to get disconnected animals
//    - we multiply parts who's spin can be longer then 1/2
// old -> see rekursiveConfigs
// JS0912: melt with findExcitedFundamentalSpinConfigurations
//  diffs: int NumberConfigurations - NumberExcitedConfigurations
//         int subsetClusters(,,0)  - subsetExcitedClusters(,,1)
//         struct spinConfigs *spinConfigurations   - excitedSpinConfigurations
//         "numberConfigurations" - "numberExcitedConfigurations"
void findSpinConfigurations(struct spinConfigs *cfgs, int *nCfgs, int id)
{

  int j, l, animal, is_minDescriptor;
  long int newNumClusters=0,  allClusters=0;
  // int max_num_spin_flips;
  int descriptor[DESCR_SIZE];
  int descriptorIndex;
  int length;
  int (*configuration)[5];
  static int spin[MAX_LSUBM+1];  // ToDo: meaning LSUBN+1
  int (*latticeAnimal)[5];
  int tempLength;
  int flag, maxNumberFlips;
  static int n_out=0;
  time_t t1, t2;

  time(&t1);        /* get unix time in seconds */
//  static int configuration[MAX_LSUBM+1][5];
  configuration=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));
  latticeAnimal=(int (*)[5])malloc((MAX_LSUBM+1)*5*sizeof(int));

  // for 1D spin-one this is simple.
  
  // max_num_spin_flips = Params.s2 + 1; // 2*s+1
  
  (*nCfgs)=0;
  
  for(animal=1;animal<=NumberLatticeAnimals;animal++) {
    
    // see bookKeeping.C
    length=findClusterFromDescriptor(latticeAnimals[animal].descriptor,
				     latticeAnimal);

#if (!defined(DSUB)) && (!defined(LDSUB)) && (!defined(LPSUB))
//    if (length<Params.m) continue;  // do not work for spin=10001 trick 0912
#endif
    // shorter configs build from longest animals, Joerg Nov05
    // animal<100 added for SUB-2-10 case (very slow, lot of animals)
    if (Control.output > 3 && myid == 0 && n_out<=100 && animal<100) {
      printf("\nfindCluster: use  Animal[%d]: ", animal);
      printOutDescriptor(latticeAnimals[animal].descriptor);
      printOutConfig(latticeAnimal, length);
    }
    // if (length >  Params.n) continue; // ToDo: speedup using discAnimals

    // Joerg Oct05: we skip small clusters and build shorter from longest
    //     ToDo:    later we dont skip and dont shorten animals
    // if (length <  Params.m) continue;

    if (length > Params.m) { // ToDo: for xxzCubic-8-8 Joerg Sep05, new=.n ?
      printf("ERROR findSpinConfigurations(%d):"
             " length=%d != Params.m=%d for LatticeAnimal=%d"
             "\n", id, length, Params.m, animal);
      printOutConfig(latticeAnimal, length);
      Control.output=5;
      length=findClusterFromDescriptor(latticeAnimals[animal].descriptor,
                                     latticeAnimal);
      ERR(1);
    }

    // slowest part of the program? Joerg
    // loop over (2s+1)^(m)-1 spin states for connected animals
    //  (ToDo: skip all codes having 0 in it, also disconnected animals)
    // set starting spin code to old=00000 new=11111:
    { int i1; for(i1=0; i1<length; i1++) spin[i1+1]=0; }
    int tmp_n=0; // num flips

    for (;;) {  // loop over (2s+1)^(m)-1 spin states
      for(j=0; j<length; j++) {   // or length for animals length<m
        tmp_n-=spin[j+1];
        spin[j+1]=(spin[j+1]+1)%(Params.s2+1);
        tmp_n+=spin[j+1];
        if (tmp_n>Params.n) { j--; continue; } // skip tmp_n>n
        if (spin[j+1]) break; // continue only on overflow
      } if (j==length) break; // counting finished
      // printf(" %d%d%d%d%d..%d\n",
      //  spin[1],spin[2],spin[3],spin[4],spin[5],spin[Params.m]);

      // spin[1..m]: 001 002 010 011 012 020 021 022 100 ... for s=1
      // ToDo: this is to much! we should split n into (m=length)+1 summands
      //      filtered out in inner functions? (2s)^m -> (m over 2s)?
      // n=2 spin[1..m]: 001 002 010 011 020 100 101 110 200
      //    or           001 010 100 002 011 020 101 110 200

      // ToDo: first and last site must be a spin (flipped)
      //       so we count nothing twice  (JS0912)
      // this is a test, faster if we set first and last 1 and count n-2 bits
//      if (spin[1]==0 || spin[length]==0) continue;
//      if (spin[length]==0) continue;

      tempLength=0;
      flag=0;

      for(j=1;j<=length;j++) {
        maxNumberFlips= UnitCell.s2[latticeAnimal[j][4]];
        if (spin[j] > maxNumberFlips) { flag=1; break; } // ignore
        for(l=1;l<=spin[j];l++) { // copy this config 0..2s times
          tempLength++;
          if (tempLength > MAX_LSUBM) continue; // Overflow Error below
          // ToDo: does it make sence? makes program slow
          configuration[tempLength][1] = latticeAnimal[j][1];
          configuration[tempLength][2] = latticeAnimal[j][2];
          configuration[tempLength][3] = latticeAnimal[j][3];
          configuration[tempLength][4] = latticeAnimal[j][4];
        }
      }

      if (flag) continue;

      if (tempLength > MAX_LSUBM) {
        printf("Error findSpinConfigurations(%d):"
               " tempLength=%d > MAX_LSUBM\n", id, tempLength);
        ERR(1);
      }
              
      // ToDo: better call Subset before calling  slow functions!
      //   call subsetClusters with cluster instead of descriptor!
      if (!subsetClusters(configuration,tempLength,id)) continue;

      // ToDo Joerg: is it necessary to check all symmetrized
      //             configs here? Its slow! 
      //         if animals are symmetrized, we can skip symmetry here
      // get descriptor from symmetrized configuration[tempLength]
      //  add 1 if we have already the minimum descriptor
  // ToDo: problems on NumberConfigurations MAX_NUM_CONFIGURATIONS-overflow
      is_minDescriptor=
        findDescriptorOfCluster(tempLength, configuration, descriptor);

      if (is_minDescriptor<0) ERR(999999); // continue??? Joerg Oct09
      // ensure that we dont search outside the array on overflow JS0912
      { int tmpNC=(*nCfgs);
        if (tmpNC>MAX_NUM_CONFIGURATIONS)
            tmpNC=MAX_NUM_CONFIGURATIONS;
        descriptorIndex = findDescriptorIndex(descriptor, cfgs, tmpNC);
      }

//      if (is_minDescriptor == 1 && !descriptorIndex) newNumClusters++;
      if (is_minDescriptor == 1) newNumClusters++; // JS0912
      allClusters++;   // count all checked clusters (measure performance)
      
      if (Control.output > 2 && myid == 0 && n_out<=100) {
        printf("\n fC-test Animal %d spins=", animal);
        { int i1; for (i1=0;i1<length;i1++) printf("%d",spin[i1+1]); }
        // printOutConfig(configuration, tempLength);
        printOutConfig(latticeAnimal, length); // more DGB info
      }
      if (Control.output > 2 && myid == 0 && n_out<=100) {
        if (Control.output > 3) {
          printf("transformed into: ");
          printOutClusterFromDescriptor(descriptor);
        }
// JS0912
        if  (is_minDescriptor==1 && descriptorIndex)
          printf(" Warning: newNumClusters count this config too much\n");
        printf(" newNumCfgs= %ld  tests= %ld  minD= %d",
            newNumClusters, allClusters, is_minDescriptor);  // testwise
        printf(" stored= %d", descriptorIndex); // 0 means "not yet stored"
        printf(" subset= %d\n",subsetClusters(configuration,tempLength,id));
      }
      if (n_out<100000) n_out++; // output counter
      
      // also filter out here length > Params.n (spin flips)
      if (!descriptorIndex) {
        static long nc_out = 1;
        updateDescriptors(descriptor,cfgs,(*nCfgs));
        (*nCfgs)++;
        if ((*nCfgs)>=MAX_NUM_CONFIGURATIONS
         || (*nCfgs)<0) { ERR(999);exit(1); } 

        if (!myid)        
        if ((Control.output>2 && (*nCfgs)<=100)
         || (Control.output>0 && (*nCfgs)==nc_out)) {
          if ((*nCfgs)==nc_out)
             if (nc_out<1<<16) nc_out<<=1; else nc_out+=1<<16;
          printf("add config = %6d: ", (*nCfgs));
          printOutClusterFromDescriptor(descriptor);
          if ((*nCfgs)==100) printf("... skip output ...");
        }
      }
    }
  }
  time(&t2);        /* get unix time in seconds */

  // counted without lookup to the configuration table (Joerg Nov05)
  if(id==0 && !myid) printf("numberFundamentalConfigurations = %ld\n", (long)(*nCfgs));
  if(id==1 && !myid) printf("numberExcitedConfigurations = %ld\n", (long)(*nCfgs));
  if(!myid)
    printf("newNumCfgs= %ld : allCfgs = %ld : time[min]= %.2f\n",
            newNumClusters, allClusters, (t2-t1)/60.);

#if EXPERIMENTAL & 1  // defined in Makefile
  t1=t2; j=0;
  // try new method
  // ToDo: split for Fundamental an Excited!
  if (id==0) j=findRecursiveConfigs(0, configuration); /* args: len, cfg */
  if (id==1) printf("Warning: recursiveF() is bad for excitations (ToDo)\n");
  time(&t2);        /* get unix time in seconds */
  if (!myid)
    printf("findRecursiveConfigs nC= %d  time[min]= %.2f\n",j,(t2-t1)/60.);
#else
  if (!myid) printf("findRecursiveConfigs skipped, !(EXPERIMENTAL & 1)\n");
#endif

  free(configuration);
  free(latticeAnimal);
  if ((*nCfgs) > MAX_NUM_CONFIGURATIONS) {
    printf("ERROR: MAX_NUM_CONFIGURATIONS excited, NumCfgs= %d\n", (*nCfgs));
    ERR(1);
  }
}


// Jun08 Joerg ToDo: melt with updateExcitedDescriptors
// dTable: spinConfigurations or excitedSpinConfigurations or latticeAnimals
// Nov09: add dTable
// Jun08 Joerg  ToDo: melt with updateDescriptors
//    only differs in ...
// void updateExcitedDescriptors(int *descriptor, int numberElements) ...
// ToDo: add a hashing table? to make it faster?
// ToDo: search table and insert if not found, return if found
// ToDo: start,len table for each length n (sorted by n)
// void updateDescriptors2(int *descriptor, int numberElements) ...
void updateDescriptors(int *descriptor,
   struct spinConfigs *dTable, int numberElements)
{
  int i, position;
  
  // first find where to put the data element 
  // just use a simple backwards linear search

  // although this bit is as cumbersome as old routine it is called
  // less often and all of the elements are always sorted!

  if (numberElements >= MAX_NUM_CONFIGURATIONS-1
   || numberElements < 0) {
      if (!myid)
        printf("\nERROR updateDescr: counter overflow (%d, %d)\n",
               numberElements, MAX_NUM_CONFIGURATIONS);
      ERR(__LINE__); exit(1);
  }

  position=1;

  if (numberElements > 0) {
#if 0
    int tempLogicFlag, j;
    for(j=numberElements;j>0;j--) {
      tempLogicFlag=cmpDescriptors(descriptor, dTable[j].descriptor);
      if (tempLogicFlag==1) { position=j+1; break; }
#if 0
      dTable[j+1]=dTable[j  ]
#endif
    }
#else
  int t1, t2, middle;
  t1=1-1;
  t2=numberElements+1;
  do {
    middle=(t1+t2)/2; // if t1+t2 is odd 
    for(i=DESCR_SIZE-1;i>=0;i--) {
      if (dTable[middle].descriptor[i]==descriptor[i]) continue;
      if (dTable[middle].descriptor[i] >descriptor[i]) 
        t2=middle; // new upper bound
      else
        t1=middle; // new lower bound
      break;
    }      
    if (i<0) { ERR(234); exit(1); } // found it!?
  } while (t2-t1 > 1);
  position=t2;	       
#endif

#if 0
  for(i=numberElements;i>=position;i--) dTable[i+1]=dTable[i];
#else
  // dramatic speedup (x25) for transfers outside cpu_cache
  memmove(&dTable[position+1],  // dest
          &dTable[position  ],  // src
         (numberElements-position+1)*sizeof(struct spinConfigs));
#endif
  } // numberElements > 0
 
  for(i=0;i<DESCR_SIZE;i++)
    dTable[position].descriptor[i]=descriptor[i];
  
#if DEBUG
{
  static long n_out=0, n_out2=1;
  n_out++;
  if (Control.output > 3 && n_out==n_out2 && !myid) {
    int j;
    n_out2<<=1;
    printf("updateDescr: print table ...\n");
    for(j=1;j<=numberElements+1;j++)
    if (j<4 || j>numberElements-3) {
      printf(" descr[%6d]: ", j);
      for(i=0;i<DESCR_SIZE;i++) printf("%08x ", dTable[j].descriptor[i]);
      printf("\n");
    }
    printf("\n");
  }
}
#endif
}


// filter out unwished states (return flag=0 for throughing away)
// called by findFundamentalSpinConfigurations()
// Oct05: use cluster + length instead of descriptor as argument!
//    config[i]={0,x,y,z,uc}
// Jun08 Joerg ToDo: melt subset[Excited]Clusters if possible
//  ToDo: move to sample_....C file! #ifndef subSet...
// id: 0=FundamentalConfigs 1=ExcitedConfigs JS0912
int subsetClusters(int (*config)[5], int length, int id)
{

  int i;

  // this routine is somewhat lattice specific and is used to 
  // restrict the clusters used to within some well-defined
  // subset.

  // Hence, this subroutine is application-dependent and to be 
  // written by the user.

  // it returns a value of 0 if the cluster is not acceptable
  // and a non-zero integer otherwise.

  // always check that we don't have more than n spin flips.
  if (length > Params.n) return 0; 

  // restriction 1: NO RESTRICTION!

  // restriction 2: even number of spin flips.
  if (restrictionClusters==2 && id==0 &&   length&1 ) return 0;
  if (restrictionClusters==2 && id==1 && !(length&1)) return 0; // JS091021
  
  // restriction 3: S_T^z=0
  if (restrictionClusters==3) {

    int stz;  // Apr08 JS: stz from double to int, better code

    // USER DEFINED OPTION!!

    // ADVANCED: USER MUST ALTER CODE IN THIS CASE!!!

    // EXAMPLE: XXZ model or Heisenberg linear-chain, square or cubic
    // lattices arbitrary `s' though. Neel state with same number of
    // spin flips on A- and B-sites.

    stz=0;
    for(i=1;i<=length;i++) {
// abbreviations to get more compact readable subset definitions
// RX,RY,RZ are the integer units of the unitcell
// UC is the site number within unitcell 1..nUC (Joerg Sep09)
// ToDo: shift the A,B sublattice definition in CCM input file
#define UC config[i][4]
#define RX config[i][1]
#define RY config[i][2]
#define RZ config[i][3]
// ToDo Joerg09: switching here at source level often caused bad computations
#ifndef Subset3Rule
//  we need one set of Subset* defined here
#define Subset3Rule 1 & (RX+RY+RZ)
#define Subset3Out "1 & (rx+ry+rz)  for simple square, cubic, ..."
//#define Subset3Rule 1 & (RX)
//#define Subset3Out "1 & (rx)  for striped models xA=0,2,4,... xB=1,3,5,..."
//#define Subset3Rule 1 & (UC)
//#define Subset3Out "1 & (uc)  for SrCuBO, square_ext_field, staggered dimer, LC_N_6"
//#define Subset3Rule 1 & ((   0x0c>>UC)+RX+RY)
//#define Subset3Out "1 & ((0b01100>>uc)+rx+ry)  for CaVO-Neel(UC reindexed)"
//#define Subset3Rule 1 & (   0x0c>>UC)  
//#define Subset3Out "1 & (0b01100>>uc)  for SrCuBO-Stripe(UC reindexed)"  
#endif
      if ( Subset3Rule )
        stz++;
      else
        stz--;
    }
#if DEBUG
    if (Control.output>6)
        printf("subsetClusters: stz= %4d flag= %d\n", stz, ((stz)?0:1));
#endif
    if (id==0 && stz) return 0;
    if (id==1 && stz!=1 && stz!=-1) return 0;
  }

  // new restriction, especially for canted spin states. 9-8-07 DJJF
  // JS: canted states (angle) and length=1 clusters cause solve.C to fail
  if (id==0 && restrictionClusters==4 && length==1) return 0;
// if (id==1 && restrictionClusters==4 && length==1) return 0;

  return 1;
}
// see subsetClusters() above for comments
// int subsetExcitedClusters() replaced by subsetClusters(,,1) JS0912

// Subset3Out must be identical to "Subset3Rule"
// print out subset rule for debugging
void subsetPrint( void ) {
  printf(" subsetPrint: restriction=%d\n", restrictionClusters);
  printf("  1=no restriction, 2=even flips, 3=user defined, 4=lenght>1\n");
  printf("  user defined Subset3Rule: %s\n", Subset3Out);
  printf("  please verify with grep define\\ Subset define.C\n");
}
